local __exports = LibStub:NewLibrary("ovale/states/soulbind", 90105)
if not __exports then return end
local __class = LibStub:GetLibrary("tslib").newClass
local aceEvent = LibStub:GetLibrary("AceEvent-3.0", true)
local aceTimer = LibStub:GetLibrary("AceTimer-3.0", true)
local ipairs = ipairs
local pairs = pairs
local unpack = unpack
local concat = table.concat
local insert = table.insert
local sort = table.sort
local C_Soulbinds = C_Soulbinds
local Enum = Enum
local GetSpellInfo = GetSpellInfo
local __enginecondition = LibStub:GetLibrary("ovale/engine/condition")
local returnBoolean = __enginecondition.returnBoolean
local returnConstant = __enginecondition.returnConstant
local __enginedbc = LibStub:GetLibrary("ovale/engine/dbc")
local conduits = __enginedbc.conduits
__exports.Soulbind = __class(nil, {
    constructor = function(self, ovale, debug)
        self.conduitSpellIdById = {}
        self.conduitId = {}
        self.conduitRank = {}
        self.isActiveConduit = {}
        self.isActiveTrait = {}
        self.debugConduitOptions = {
            type = "group",
            name = "Conduits",
            args = {
                conduits = {
                    type = "input",
                    name = "Conduits",
                    multiline = 25,
                    width = "full",
                    get = function()
                        return self.debugConduits()
                    end
                }
            }
        }
        self.debugSoulbindTraitsOptions = {
            type = "group",
            name = "Soulbind Traits",
            args = {
                soulbindTraits = {
                    type = "input",
                    name = "Soulbind Traits",
                    multiline = 25,
                    width = "full",
                    get = function()
                        return self.debugSoulbindTraits()
                    end
                }
            }
        }
        self.onEnable = function()
            self.module:RegisterEvent("COVENANT_CHOSEN", self.onSoulbindDataUpdated)
            self.module:RegisterEvent("PLAYER_ENTERING_WORLD", self.onSoulbindDataUpdated)
            self.module:RegisterEvent("PLAYER_LOGIN", self.onPlayerLogin)
            self.module:RegisterEvent("SOULBIND_ACTIVATED", self.onSoulbindDataUpdated)
            self.module:RegisterEvent("SOULBIND_CONDUIT_COLLECTION_CLEARED", self.onSoulbindConduitCollectionUpdated)
            self.module:RegisterEvent("SOULBIND_CONDUIT_COLLECTION_REMOVED", self.onSoulbindConduitCollectionUpdated)
            self.module:RegisterEvent("SOULBIND_CONDUIT_COLLECTION_UPDATED", self.onSoulbindConduitCollectionUpdated)
            self.module:RegisterEvent("SOULBIND_CONDUIT_INSTALLED", self.onSoulbindDataUpdated)
            self.module:RegisterEvent("SOULBIND_CONDUIT_UNINSTALLED", self.onSoulbindDataUpdated)
            self.module:RegisterEvent("SOULBIND_NODE_LEARNED", self.onSoulbindDataUpdated)
            self.module:RegisterEvent("SOULBIND_NODE_UNLEARNED", self.onSoulbindDataUpdated)
            self.module:RegisterEvent("SOULBIND_NODE_UPDATED", self.onSoulbindDataUpdated)
            self.module:RegisterEvent("SOULBIND_PATH_CHANGED", self.onSoulbindDataUpdated)
            self.onSoulbindConduitCollectionUpdated("onEnable")
        end
        self.onDisable = function()
            self.module:UnregisterEvent("COVENANT_CHOSEN")
            self.module:UnregisterEvent("PLAYER_ENTERING_WORLD")
            self.module:UnregisterEvent("PLAYER_LOGIN")
            self.module:UnregisterEvent("SOULBIND_ACTIVATED")
            self.module:UnregisterEvent("SOULBIND_CONDUIT_COLLECTION_CLEARED")
            self.module:UnregisterEvent("SOULBIND_CONDUIT_COLLECTION_REMOVED")
            self.module:UnregisterEvent("SOULBIND_CONDUIT_COLLECTION_UPDATED")
            self.module:UnregisterEvent("SOULBIND_CONDUIT_INSTALLED")
            self.module:UnregisterEvent("SOULBIND_CONDUIT_UNINSTALLED")
            self.module:UnregisterEvent("SOULBIND_NODE_LEARNED")
            self.module:UnregisterEvent("SOULBIND_NODE_UNLEARNED")
            self.module:UnregisterEvent("SOULBIND_NODE_UPDATED")
            self.module:UnregisterEvent("SOULBIND_PATH_CHANGED")
        end
        self.onPlayerLogin = function(event)
            self.module:ScheduleTimer(self.onPlayerLoginDelayed, 3)
        end
        self.onPlayerLoginDelayed = function()
            local event = "PLAYER_LOGIN"
            self.tracer:debug(event)
            self.onSoulbindConduitCollectionUpdated(event)
            self.onSoulbindDataUpdated(event)
        end
        self.onSoulbindConduitCollectionUpdated = function(event)
            self.tracer:debug(event .. ": Updating conduit collection.")
            for _, conduitType in pairs(Enum.SoulbindConduitType) do
                local collectionData = C_Soulbinds.GetConduitCollection(conduitType)
                for _, data in ipairs(collectionData) do
                    local id = data.conduitID
                    local rank = data.conduitRank
                    local spellId = C_Soulbinds.GetConduitSpellID(id, rank)
                    self.conduitSpellIdById[id] = spellId
                    self.conduitId[spellId] = id
                    self.conduitRank[spellId] = rank
                end
            end
        end
        self.onSoulbindDataUpdated = function(event)
            self.tracer:debug(event .. ": Updating soulbind data.")
            self.isActiveConduit = {}
            self.isActiveTrait = {}
            local soulbindId = C_Soulbinds.GetActiveSoulbindID() or 0
            self.tracer:debug(event .. ": active soulbind = " .. soulbindId)
            if soulbindId ~= 0 then
                local data = C_Soulbinds.GetSoulbindData(soulbindId)
                for _, node in pairs(data.tree.nodes) do
                    local isSelected = node.state == Enum.SoulbindNodeState.Selected
                    local isTrait = node.spellID and node.spellID ~= 0
                    local isConduit = node.conduitID and node.conduitID ~= 0
                    local isActiveConduit = isConduit and isSelected
                    if isTrait then
                        self.isActiveTrait[node.spellID] = isSelected
                    end
                    if isActiveConduit then
                        local spellId = self.conduitSpellIdById[node.conduitID]
                        self.isActiveConduit[spellId] = true
                    end
                end
            end
        end
        self.debugConduits = function()
            local output = {}
            for spellId, id in pairs(self.conduitId) do
                local rank = self.conduitRank[spellId]
                local name = GetSpellInfo(spellId)
                if self.isActiveConduit[spellId] then
                    insert(output, name .. ": " .. spellId .. ", id=" .. id .. ", rank=" .. rank .. " (active)")
                else
                    insert(output, name .. ": " .. spellId .. ", id=" .. id .. ", rank=" .. rank)
                end
            end
            sort(output)
            return concat(output, "\n")
        end
        self.debugSoulbindTraits = function()
            local output = {}
            for spellId, isActive in pairs(self.isActiveTrait) do
                local name = GetSpellInfo(spellId)
                if isActive then
                    insert(output, name .. ": " .. spellId .. " (active)")
                else
                    insert(output, name .. ": " .. spellId)
                end
            end
            sort(output)
            return concat(output, "\n")
        end
        self.conduitCondition = function(positionalParameters)
            local id = unpack(positionalParameters)
            local conduitId = self.conduitId[id] or id
            local spellId = self.conduitSpellIdById[conduitId]
            return returnBoolean(self.isActiveConduit[spellId])
        end
        self.conduitRankCondition = function(positionalParameters)
            local id = unpack(positionalParameters)
            local conduitId = self.conduitId[id] or id
            local spellId = self.conduitSpellIdById[conduitId]
            local rank = self.conduitRank[spellId]
            if rank then
                return returnConstant(rank)
            else
                return 
            end
        end
        self.soulbindCondition = function(positionalParameters)
            local spellId = unpack(positionalParameters)
            return returnBoolean(self.isActiveTrait[spellId])
        end
        self.conduitValue = function(atTime, id)
            local conduitId = self.conduitId[id] or id
            local spellId = self.conduitSpellIdById[conduitId]
            local rank = self.conduitRank[spellId]
            if rank then
                local value = conduits[conduitId].ranks[rank]
                return returnConstant(value)
            else
                return 
            end
        end
        debug.defaultOptions.args["conduit"] = self.debugConduitOptions
        debug.defaultOptions.args["soulbindTraits"] = self.debugSoulbindTraitsOptions
        self.module = ovale:createModule("Soulbind", self.onEnable, self.onDisable, aceEvent, aceTimer)
        self.tracer = debug:create(self.module:GetName())
    end,
    registerConditions = function(self, condition)
        condition:registerCondition("conduit", false, self.conduitCondition)
        condition:registerCondition("conduitrank", false, self.conduitRankCondition)
        condition:registerCondition("enabledsoulbind", false, self.soulbindCondition)
        condition:registerCondition("soulbind", false, self.soulbindCondition)
        condition:register("conduitvalue", self.conduitValue, {
            type = "number"
        }, {
            name = "conduit",
            type = "number",
            optional = false
        })
    end,
})
